! SAOXML5RW.f90 - reading and writing interface for SAO.XML format in Fortran
!
! Version: v 0.4 2010/11/02
!
! Ivan Galkin
! University of Massachusetts Lowell
! Center for Atmospheric Research
! Email: Ivan_Galkin@uml.edu
!
! Edited 2010/11/2 by Mark A. Hausman for additional error-handling.
!  Also edited subroutine saoxml_write_system_info (removed second write of FrequencyStepping end tag)
!
! General information:
! SAO.XML is standard arrhival output format for ionogram-derived data
! Version 5 was designed in 2006-2007
! For more information see http://umlcar.uml.edu/SAOXML/
!
! Reading code fills structure IonogramDerivedData from file
! Writing code outputs contents of IonogramDerivedData to file
!
! The SAOXML5RW library uses Fortran library for XML operations by Arjen Markus, XML-Fortran
!
module SAOXML5RW
   use IonogramDerivedData
   use XmlParse
   use read_xml_primitives
   implicit none

   integer                                :: major_version, minor_version ! in the input data
   
   integer,parameter                      :: MAJOR_VERSION_OUTPUT = 5 ! for the output data
   integer,parameter                      :: MINOR_VERSION_OUTPUT = 0 ! for the output data

   logical                                :: mustread
   type(XML_PARSE)                        :: info
   character(len=80)                      :: tag
   logical                                :: endtag
   character(len=80),dimension(1:2,1:20)  :: attribs
   integer                                :: no_attribs
   character(len=2000),dimension(1:100)   :: data
   integer                                :: no_data
   logical                                :: has_var


   ! GLOBAL VARIABLES FOR VARIOUS STATE MACHINES
   
   character(len=IDD_MAX_VALUE_NAME),private :: trace_value_name = ''
   character(len=IDD_MAX_VALUE_NAME),private :: profile_value_name = ''
   real(8),private                           :: no_value_trace
   real(8),private                           :: no_value_profile
   type(IDD_PROFILE_COEFS),private           :: tmp_cheb
   type(IDD_QP_SEGMENT),private              :: tmp_qp
   integer,private                           :: currently_reading 
   integer,parameter,private                 :: FREQUENCY_STEPPING = 1
   integer,parameter,private                 :: RANGE_STEPPING = 2
   logical                                   :: reading_contact_person

   ! other commonly used private variables
   integer, private                          :: stat ! to capture errors of reading empty fields
   integer, private                          :: i
   character(len=1000), private              :: buffer
   character(len=20), private                :: cbuf
!
   logical                                   :: profile_bounds_available
   logical                                   :: profile_lower_bounds_available
   logical                                   :: profile_upper_bounds_available
!
!  options for output content
   logical                                   :: optional_density_profile_output = .false.

!   private :: read_attributes
!   private :: read_data
!   private :: read_saorecord_attributes
!   private :: read_systeminfo_attributes
!   private :: read_one_standard_characteristic
!   private :: read_one_modeled_characteristic
!   private :: read_one_custom_characteristic
!   private :: start_reading_one_trace
!   private :: identify_trace_value
!   private :: read_trace_frequencies
!   private :: read_trace_ranges
!   private :: read_trace_values
!
!   interface read_from_buffer
!      module procedure saoxml_open_input_file
!      module procedure saoxml_scan_input
!      module procedure saoxml_close_input_file
!      module procedure saoxml_read_one_record
!   end interface

contains

!###################################################################################################
!
!
!      S   A   O   X   M   L       R   E   A   D   I   N   G       F   U   N   C   T   I   O   N   S
!
!
!###################################################################################################

!
!---   s a o x m l _ o p e n _ i n p u t _ f i l e  ( )    ----------------------------------------
!     Calling routine should check for successful return with a call to xml_ok(info)
!
subroutine saoxml_open_input_file( fname )

    character(len=*) :: fname

! Assign the xml-filename to fname and open the file
   mustread = .true.
   call xml_open(info,fname,mustread)
! Check for errors
   if( .not. xml_ok(info) ) then  !  eof, IO error, or illegal content;  stop reading
! if file was successfully opened (signified by info%lun > 0), but had illegal content,
!  close the file before proceeding
      if( info%lun > 0) call saoxml_close_input_file()
      return
   else
      call xml_options(info,ignore_whitespace = .true.)
   endif
end subroutine saoxml_open_input_file

!
!---   s a o x m l _ c l o s e _ i n p u t _ f i l e  ( )       -----------------------------------
!
subroutine saoxml_close_input_file()
   logical opend
!   make sure file is actually open before attempting to close it (avoid run-time Fortran errors)
   if( info%lun <= 0) return
   inquire( unit = info%lun, opened = opend )
   if ( opend ) call xml_close(info)
   return
end subroutine saoxml_close_input_file

!
!---   s a o x m l _ s c a n _ i n p u t ( )     --------------------------------------------------
!
! Scan the file, count how many SAORecords provided, rewind
! 
! Return: number of SAORecords in the input file
!
integer function saoxml_scan_input()
   saoxml_scan_input = 0
   do
      call xml_get(info,tag,endtag,attribs,no_attribs,data,no_data)
      if (xml_error(info)) then
      ! ToDo: handle the errors
      endif
      ! simply count number of <SAORecord> opening tags
      if(   tag .eq. TAG_SAO_RECORD .and. .not. endtag ) then
         saoxml_scan_input = saoxml_scan_input + 1
      ! stop at closing tag, do not let read beyond EOF
      endif
      if(   tag .eq. TAG_SAO_RECORD_LIST .and. endtag ) exit
      if( .not. xml_ok(info) ) exit
   enddo

   rewind( info%lun )
   tag = '' ! clear the tag for the subsequent read operations in the caller
   info%eof = .false. ! clear eof condition from scanning the file to the end
end function saoxml_scan_input

!
!---   s a o x m l _ r e a d _ o n e _ r e c o r d ( )      ---------------------------------------    
!     Calling routine should check for successful return with a call to xml_ok(info)
!
subroutine saoxml_read_one_record()

! skip file to the closest <SAORecord> tag
   do
      call xml_get(info,tag,endtag,attribs,no_attribs,data,no_data)
   ! ToDo: handle the errors
      if( .not. xml_ok(info) ) then  !  eof, IO error, or illegal content;  stop reading
         return
      endif
      if( tag .eq. TAG_SAO_RECORD .and. .not.endtag ) exit
   enddo

! reset IDD data structure that will hold the record
   call idd_reset()

! process SAORecord attributes
   call read_saorecord_attributes()

! keep reading the record until </SAORecord> tag is found
   do
      call xml_get(info,tag,endtag,attribs,no_attribs,data,no_data)
   ! ToDo: handle the errors
      if( .not. xml_ok(info) ) then  !  eof, IO error, or illegal content;  stop reading
         return
      endif
      if( no_attribs > 0 ) call read_attributes()
      if( no_data > 0 ) call read_data()

      ! exit when the record is over
      if( tag .eq. TAG_SAO_RECORD .and. endtag ) exit
      
      ! some elements require support on the higher level of element hierarchy:

      ! make sure <Name> tag belongs to person, not something else
      if( tag .eq. TAG_CONTACT_PERSON) reading_contact_person = .true.
      
      ! when profile value element closes, fill missing bounds
      if( tag .eq. TAG_PROFILE_VALUE_LIST .and. endtag) call finalize_profile_value_uncertainties()
      ! when profile tabulation element closes, makes sure both density and frequency have bounds
      if( tag .eq. TAG_TABULATED .and. endtag) call finalize_profile_uncertainties()
      if( .not. xml_ok(info) ) exit
   enddo
end subroutine saoxml_read_one_record


!
!---   r e a d _ a t t r i b u t e s ( )    -------------------------------------------------------
!
!
! read_attributes() is called when an opening tag is found and 
! the attributes for the XMl element become available
!
! read_attributes() works as a dispatcher by calling appropriate 
! reading subroutines for a specific element
!
subroutine read_attributes()

   select case ( tag )
   case (TAG_SAO_RECORD)
      call read_saorecord_attributes()
   case (TAG_SYSTEM_INFO)
      call read_systeminfo_attributes()
   case (TAG_FREQUENCY_STEPPING)
      call start_reading_frequency_stepping()
   case (TAG_RANGE_STEPPING)
      call start_reading_range_stepping()
   case (TAG_LINEAR_STEPPING)
      call read_linear_stepping()
   case (TAG_LOG_STEPPING)
      call read_log_stepping()
   case (TAG_RESTRICTED_FREQUENCY_LIST)
      call start_reading_rfil()
   case (TAG_AUTO_SCALER)
      call read_auto_scaler_info()
   case (TAG_MANUAL_SCALER)
      call read_manual_scaler_info()
   case (TAG_GYROFREQUENCY)
      call read_gyrofrequency()
   case (TAG_DIP_ANGLE)
      call read_dip_angle()
   case (TAG_SUNSPOT_NUMBER)
      call read_sunspot_number()
!  case (TAG_CHARACTERISTIC_LIST)
   case (TAG_URSI)
      call read_one_standard_characteristic()
   case (TAG_MODELED)
      call read_one_modeled_characteristic()
   case (TAG_CUSTOM)
      call read_one_custom_characteristic()
!  case (TAG_TRACE_LIST)
   case (TAG_TRACE)
      call start_reading_one_trace()
!  case (TAG_FREQUENCY_LIST)
!  case (TAG_RANGE_LIST)
   case (TAG_TRACE_VALUE_LIST)
      call identify_trace_value()
!  case (TAG_PROFILE_LIST)
   case (TAG_PROFILE)
      call start_reading_one_profile()
   case (TAG_TABULATED)
      call start_reading_one_tabulated_profile()
   case (TAG_PROFILE_VALUE_LIST)
      call identify_profile_value()
   case (TAG_SHIFTED_CHEBYSHEV)
      call read_profile_shifted_chebyshev_attributes()
   case (TAG_QUASI_PARABOLIC_LIST)
      call start_reading_qp_segments()
   case (TAG_QUASI_PARABOLIC)
      call read_one_qp_segment()
   case (TAG_PROFILE_VALLEY)
      call read_profile_valley()
   case (TAG_TOPSIDE_CHAPMAN)
      call read_profile_topside_chapman_layer()
   end select

end subroutine read_attributes

!
!---   r e a d _ d a t a ( )    -------------------------------------------------------
!
!
! read_data() is called when element data are acquired completely
! this may be several lines of text in data array (no_data lines)
!
! read_data() works as a dispatcher by calling appropriate 
! reading subroutines for a specific element
!
subroutine read_data()
   select case ( tag )
   case (TAG_COMMENTS)
      call read_comments()
   case (TAG_FREQUENCY_LIST)
      call read_trace_frequencies()
   case (TAG_LOWER_LIMIT_LIST)
      call read_lower_restricted_limits()
   case (TAG_UPPER_LIMIT_LIST)
      call read_upper_restricted_limits()
   case (TAG_NAME)
      call read_contact_person_name()
   case (TAG_AFFILIATION)
      call read_contact_person_affiliation()
   case (TAG_ADDRESS)
      call read_contact_person_address()
   case (TAG_EMAIL)
      call read_contact_person_email()
   case (TAG_DIGISONDE_PREFACE)
      call read_digisonde_preface()
   case (TAG_RANGE_LIST)
      call read_trace_ranges()
   case (TAG_TRACE_VALUE_LIST)
      call read_trace_values()
   case (TAG_ALTITUDE_LIST)
      call read_profile_altitudes()
   case (TAG_PROFILE_VALUE_LIST)
      call read_profile_values()
   case (TAG_BOUND_LIST)
      call read_profile_bounds()
   case (TAG_LOWER_BOUND_LIST)
      call read_profile_lower_bounds()
   case (TAG_UPPER_BOUND_LIST)
      call read_profile_upper_bounds()
   case (TAG_SHIFTED_CHEBYSHEV)
      call read_profile_shifted_chebyshev_coefs()
   end select

end subroutine read_data

!
!###############      T  O  P    L  E  V  E  L    M  E  T  A  D  A  T  A          #################
!

!
!---   r e a d _ s a o r e c o r d _ a t t r i b u t e s ( )    -----------------------------------
!
!
! read_saorecord_attributes() is called when SAORecord opening tag is found 
! and SAORecord attributes become immediately available
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
! 
subroutine read_saorecord_attributes()

   integer :: k

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_FORMAT_VERSION)
         k = index( attribs(2,i), '.' )
         if( k > 0 ) then
            read( attribs(2,i)(1:k-1), * ) major_version
            read( attribs(2,i)(k+1:), * ) minor_version
         else
            read( attribs(2,i), * ) major_version
            minor_version = 0
         endif

      case (ATTR_START_TIME_UTC)
         read( attribs(2,i)(01:04), *, IOSTAT=stat ) idd%time%year
         read( attribs(2,i)(06:07), *, IOSTAT=stat ) idd%time%month
         read( attribs(2,i)(09:10), *, IOSTAT=stat ) idd%time%day_of_month
         read( attribs(2,i)(09:10), *, IOSTAT=stat ) idd%time%day_of_month
         read( attribs(2,i)(13:15), *, IOSTAT=stat ) idd%time%day_of_year
         read( attribs(2,i)(17:18), *, IOSTAT=stat ) idd%time%hour
         read( attribs(2,i)(20:21), *, IOSTAT=stat ) idd%time%minute
         read( attribs(2,i)(23:24), *, IOSTAT=stat ) idd%time%second
         read( attribs(2,i)(26:27), *, IOSTAT=stat ) idd%time%millisecond

      case (ATTR_URSI_CODE)
         idd%location%ursi_code = attribs(2,i)

      case (ATTR_STATION_NAME)
         idd%location%station_name = attribs(2,i)
      
      case (ATTR_GEO_LATITUDE)
         read( attribs(2,i), *, IOSTAT=stat ) idd%location%geo_latitude 

      case (ATTR_GEO_LONGITUDE)
         read( attribs(2,i), *, IOSTAT=stat ) idd%location%geo_longitude 

      case (ATTR_SOURCE)
         idd%system%source = attribs(2,i)
!        if( idd%system%source .ne. VAL_IONOSONDE ) write(*,*) 'ERROR: source data are not ionogram-derived'

      case (ATTR_SOURCE_TYPE)
         idd%system%source_type = attribs(2,i)

      case (ATTR_SCALER_TYPE)
         idd%system%autoscaled = attribs(2,i) .eq. VAL_AUTO

      case default 
         ! nothing... some attribute that we don't recognize

      end select
   enddo
   reading_contact_person = .false.
end subroutine read_saorecord_attributes

!
!---    r e a d i n g _ s y s t e m i n f o _ a t t r i b u t e s ( )  -----------------
!
!

subroutine read_systeminfo_attributes()

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_UML_STATION_ID)
         read( attribs(2,i), *, IOSTAT=stat ) idd%location%uml_sid 

      end select
   enddo

end subroutine read_systeminfo_attributes

!
!---    r e a d i n g _ i o n o g r a m _ m o d e _ d e s c r i p t i o n s ------------
!
!
subroutine start_reading_frequency_stepping()
   
   currently_reading = FREQUENCY_STEPPING

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_START_FREQUENCY)
         read( attribs(2,i), *, IOSTAT=stat ) idd%system%ionogram%start_frequency_MHz 

      case (ATTR_STOP_FREQUENCY)
         read( attribs(2,i), *, IOSTAT=stat ) idd%system%ionogram%stop_frequency_MHz 

      end select
   enddo
end subroutine start_reading_frequency_stepping

subroutine start_reading_range_stepping()
   
   currently_reading = RANGE_STEPPING

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_START_RANGE)
         read( attribs(2,i), *, IOSTAT=stat ) idd%system%ionogram%start_range_km

      case (ATTR_STOP_RANGE)
         read( attribs(2,i), *, IOSTAT=stat ) idd%system%ionogram%stop_range_km

      end select
   enddo
end subroutine start_reading_range_stepping

subroutine read_linear_stepping()
   if( currently_reading .eq. RANGE_STEPPING ) then
      idd%system%ionogram%range_stepping_rule = IDD_LINEAR_STEPPING
      call read_range_stepping
   endif
   if( currently_reading .eq. FREQUENCY_STEPPING ) then
      idd%system%ionogram%frequency_stepping_rule = IDD_LINEAR_STEPPING
      call read_frequency_stepping
   endif
end subroutine read_linear_stepping

subroutine read_log_stepping()
   if( currently_reading .eq. RANGE_STEPPING ) then
      idd%system%ionogram%range_stepping_rule = IDD_LOG_STEPPING
      call read_range_stepping
   endif
   if( currently_reading .eq. FREQUENCY_STEPPING ) then
      idd%system%ionogram%frequency_stepping_rule = IDD_LOG_STEPPING
      call read_frequency_stepping
   endif
end subroutine read_log_stepping

subroutine read_frequency_stepping()
   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_STEP)
         read( attribs(2,i), *, IOSTAT=stat ) idd%system%ionogram%frequency_step

      case (ATTR_UNITS)
         idd%system%ionogram%frequency_step_units = trim(attribs(2,i))

      end select
   enddo
end subroutine read_frequency_stepping

subroutine read_range_stepping()
   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_STEP)
         read( attribs(2,i), *, IOSTAT=stat ) idd%system%ionogram%range_step

      case (ATTR_UNITS)
         idd%system%ionogram%range_step_units = trim(attribs(2,i))

      end select
   enddo
end subroutine read_range_stepping

subroutine start_reading_rfil()
   
   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_NUM)
         read( attribs(2,i), *, IOSTAT=stat ) idd%system%ionogram%total_restricted_freq_bands 

      case (ATTR_UNITS)
         idd%system%ionogram%restricted_freq_units = trim(attribs(2,i))

      end select
   enddo
end subroutine start_reading_rfil

subroutine read_lower_restricted_limits()
   call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
          data, no_data, idd%system%ionogram%lower_restricted_freq_limits, has_var )
end subroutine read_lower_restricted_limits

subroutine read_upper_restricted_limits()
   call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
          data, no_data, idd%system%ionogram%upper_restricted_freq_limits, has_var )
end subroutine read_upper_restricted_limits

!
!---    r e a d i n g _ s c a l e r _ i n f o  ----------------------------------------
!
!
subroutine read_auto_scaler_info()

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_NAME)
         idd%system%auto_scaler = trim(attribs(2,i))

      case (ATTR_VERSION)
         idd%system%auto_scaler_version = trim(attribs(2,i))

      case (ATTR_ARTIST_FLAGS)
         read( attribs(2,i), '(20I2)', IOSTAT=stat ) idd%system%artist_flags
      end select
   enddo

end subroutine read_auto_scaler_info

subroutine read_manual_scaler_info()
   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_NAME)
         idd%system%manual_scaler = trim(attribs(2,i))

      end select
   enddo
end subroutine read_manual_scaler_info

subroutine read_comments()
   if( no_data .eq. 0 ) return
   idd%system%comments = trim(data(1))
   do i = 2,no_data
      if ( LEN(data(i)) > 0 ) &
         idd%system%comments = trim(idd%system%comments) // ' ' // trim(data(i))
   enddo
end subroutine read_comments

!
!---    r e a d  _ c o n t a c t _ p e r s o n _ i n f o  ---------------------------
!
!
subroutine read_contact_person_name()
   if( .not. reading_contact_person ) return
   idd%system%contact%name = trim( data(1) )
   idd%system%contact%available = .true.
end subroutine read_contact_person_name

subroutine read_contact_person_affiliation()
   if( .not. reading_contact_person ) return
   idd%system%contact%affiliation = trim( data(1) )
   do i = 2,no_data
      if ( LEN(data(i)) > 0 ) &
         idd%system%contact%affiliation = trim(idd%system%contact%affiliation) &
         // ' ' // trim(data(i))
   enddo
   idd%system%contact%available = .true.
end subroutine read_contact_person_affiliation

subroutine read_contact_person_address()
   if( .not. reading_contact_person ) return
   idd%system%contact%address = trim( data(1) )
   do i = 2,no_data
      if ( LEN(data(i)) > 0 ) &
         idd%system%contact%address = trim(idd%system%contact%address) &
         // ' ' // trim(data(i))
   enddo
   idd%system%contact%available = .true.
end subroutine read_contact_person_address

subroutine read_contact_person_email()
   if( .not. reading_contact_person ) return 
   idd%system%contact%email = trim( data(1) )
   idd%system%contact%available = .true.
end subroutine read_contact_person_email

!
!---   r e a d _ s o l a r _ t e r r e s t r i a l _ d a t a    ---------------------
!
!
subroutine read_gyrofrequency()
   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_VAL)
         read( attribs(2,i), *, IOSTAT=stat ) idd%geospace%gyro_frequency_MHz

      case (ATTR_MODEL)
         idd%geospace%magfield_model_name = trim(attribs(2,i))

      case (ATTR_ALTITUDE)
         read( attribs(2,i), *, IOSTAT=stat ) idd%geospace%magfield_model_altitude

      end select
   enddo
end subroutine read_gyrofrequency

subroutine read_dip_angle()
   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_VAL)
         read( attribs(2,i), *, IOSTAT=stat ) idd%geospace%dip_angle_deg

      case (ATTR_MODEL)
         idd%geospace%magfield_model_name = trim(attribs(2,i))

      case (ATTR_ALTITUDE)
         read( attribs(2,i), *, IOSTAT=stat ) idd%geospace%magfield_model_altitude

      end select
   enddo
end subroutine read_dip_angle

subroutine read_sunspot_number()
   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_VAL)
         read( attribs(2,i), *, IOSTAT=stat ) idd%geospace%sunspot_number

      case (ATTR_STATUS)
         idd%geospace%sunspot_number_status = trim(attribs(2,i))

      end select
   enddo
end subroutine read_sunspot_number

! custom system element
subroutine read_digisonde_preface()
   idd%system%digisonde_preface = trim( adjustl(data(1)) )
end subroutine read_digisonde_preface

!
!#################      C  H  A  R  A  C  T  E  R  I  S  T  I  C  S         #####################
!

!
!---   r e a d _ o n e _ s t a n d a r d _ c h a r a c t e r i s t i c ( )    ---------------------
!
!
! reads one URSI standard characteristic 
! all relevant data are available in the attributes of the <URSI> xml-element
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
subroutine read_one_standard_characteristic()

   character*2                        :: ursi_id
   real(8)                            :: value
   real(8)                            :: bound
   real(8)                            :: upper_bound
   real(8)                            :: lower_bound
   character(len=IDD_MAX_NAME)        :: boundary_type
   character                          :: ql
   character                          :: dl
   character(len=IDD_MAX_DESCRIPTION) :: description
   character(len=IDD_MAX_NAME)        :: flag

   integer     :: char_index

   ursi_id = '  '
   value = NO_VALUE_CHAR
   bound = NO_VALUE_CHAR
   upper_bound = NO_VALUE_CHAR
   lower_bound = NO_VALUE_CHAR
   boundary_type = ''
   ql = ' '
   dl = ' '
   description = ''
   flag = ''

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      ! ID is always required. Name and Units are not required and not read
      case (ATTR_ID)
         ursi_id = trim(attribs(2,i))

      ! Value is always required
      case (ATTR_VAL)
         read( attribs(2,i), *, IOSTAT=stat) value

      ! Uncertainty bound may be provided
      case (ATTR_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) bound

      ! Lower bound for the value may be provided
      case (ATTR_LOWER_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) lower_bound

      ! Upper bound for the value may be provided
      case (ATTR_UPPER_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) upper_bound

      ! Boundary type may be provided
      case (ATTR_BOUNDARY_TYPE)
         boundary_type = trim(attribs(2,i))

      ! Q letter
      case (ATTR_QL)
         ql = attribs(2,i)(1:1)

      ! D letter
      case (ATTR_DL)
         dl = attribs(2,i)(1:1)

      ! Description
      case (ATTR_UNITS)
         description = trim( attribs(2,i))

      ! Flag
      case (ATTR_FLAG)
         flag = trim( attribs(2,i))

      end select
   enddo

   char_index = idd_ursi_code_to_char_index( ursi_id )
   call idd_set_char( char_index, value )
   call idd_set_char_bounds( char_index, bound, lower_bound, upper_bound, boundary_type )
   call idd_set_char_qd_letters( char_index, ql, dl )
   if( len_trim(description) > 0 ) idd%chars(char_index)%description = trim(description)
   idd%chars(char_index)%flag = trim(flag)

end subroutine read_one_standard_characteristic

!
!---   r e a d _ o n e _ m o d e l e d _ c h a r a c t e r i s t i c ( )    ---------------------
!
!
! reads one modeled characteristic 
! all relevant data are available in the attributes of the <Modeled> xml-element
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
subroutine read_one_modeled_characteristic()

   character(len=IDD_MAX_CHAR_NAME)  :: char_name
   real(8)                           :: value
   real(8)                           :: bound
   real(8)                           :: upper_bound
   real(8)                           :: lower_bound
   character(len=IDD_MAX_NAME)       :: boundary_type
   character                         :: ql
   character                         :: dl
   character(len=IDD_MAX_UNITS_NAME) :: units
   character(len=IDD_MAX_NAME)       :: model_name
   character(len=IDD_MAX_NAME)       :: model_options
   character(len=IDD_MAX_DESCRIPTION):: description

   integer     :: char_index

   char_name = '  '
   value = NO_VALUE_CHAR
   bound = NO_VALUE_CHAR
   upper_bound = NO_VALUE_CHAR
   lower_bound = NO_VALUE_CHAR
   boundary_type = ''
   ql = '/'
   dl = '/'
   units = NO_VALUE_UNITS
   model_name = ''
   model_options = ''
   description = ''

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      ! ID is always required
      case (ATTR_NAME)
         char_name = trim(attribs(2,i))

      ! Value is always required
      case (ATTR_VAL)
         read( attribs(2,i), *, IOSTAT=stat) value

      ! Uncertainty bound may be provided
      case (ATTR_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) bound

      ! Lower bound for the value may be provided
      case (ATTR_LOWER_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) lower_bound

      ! Upper bound for the value may be provided
      case (ATTR_UPPER_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) upper_bound

      ! Boundary type may be provided
      case (ATTR_BOUNDARY_TYPE)
         boundary_type = trim(attribs(2,i))

      ! Q letter
      case (ATTR_QL)
         ql = attribs(2,i)(1:1)

      ! D letter
      case (ATTR_DL)
         dl = attribs(2,i)(1:1)

      ! Units
      case (ATTR_UNITS)
         units = trim(attribs(2,i))

      ! Model name
      case (ATTR_MODEL_NAME)
         model_name = trim(attribs(2,i))

      ! Model options
      case (ATTR_MODEL_OPTIONS)
         model_options = trim(attribs(2,i))

      ! Description
      case (ATTR_DESCRIPTION)
         description = trim( attribs(2,i))

      end select
   enddo

   char_index = idd_modeled_char_name_to_char_index( char_name )
   call idd_set_char_with_units( char_index, value, units )
   call idd_set_char_model( char_index, model_name, model_options )
   if( len_trim(description) > 0 ) idd%chars(char_index)%description = trim(description)

end subroutine read_one_modeled_characteristic

!
!---   r e a d _ o n e _ c u s t o m _ c h a r a c t e r i s t i c ( )    ---------------------
!
!
! reads one custom characteristic 
! all relevant data are available in the attributes of the <Custom> xml-element
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
subroutine read_one_custom_characteristic()

   character(len=IDD_MAX_CHAR_NAME)  :: char_name
   real(8)                           :: value
   real(8)                           :: bound
   real(8)                           :: upper_bound
   real(8)                           :: lower_bound
   character(len=IDD_MAX_NAME)       :: boundary_type
   character                         :: ql
   character                         :: dl
   character(len=IDD_MAX_UNITS_NAME) :: units
   character(len=IDD_MAX_DESCRIPTION):: description
   character(len=IDD_MAX_NAME)       :: flag

   integer     :: char_index

   char_name = '  '
   value = NO_VALUE_CHAR
   bound = NO_VALUE_CHAR
   upper_bound = NO_VALUE_CHAR
   lower_bound = NO_VALUE_CHAR
   boundary_type = ''
   ql = ' '
   dl = ' '
   flag = ''
   units = NO_VALUE_UNITS
   description = ''
   
   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      ! ID is always required
      case (ATTR_NAME)
         char_name = trim(attribs(2,i))

      ! Value is always required
      case (ATTR_VAL)
         read( attribs(2,i), *, IOSTAT=stat) value

      ! Uncertainty bound may be provided
      case (ATTR_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) bound

      ! Lower bound for the value may be provided
      case (ATTR_LOWER_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) lower_bound

      ! Upper bound for the value may be provided
      case (ATTR_UPPER_BOUND)
         read( attribs(2,i), *, IOSTAT=stat) upper_bound

      ! Boundary type may be provided
      case (ATTR_BOUNDARY_TYPE)
         boundary_type = trim(attribs(2,i))

      ! Q letter
      case (ATTR_QL)
         ql = attribs(2,i)(1:1)

      ! D letter
      case (ATTR_DL)
         dl = attribs(2,i)(1:1)

      ! Units
      case (ATTR_UNITS)
         units = trim( attribs(2,i))

      ! Description
      case (ATTR_DESCRIPTION)
         description = trim( attribs(2,i))

      ! Flag
      case (ATTR_FLAG)
         flag = trim( attribs(2,i))

      end select
   enddo

   char_index = idd_custom_char_name_to_char_index( char_name )
   call idd_set_char_with_units( char_index, value, units )
   call idd_set_char_bounds( char_index, bound, lower_bound, upper_bound, boundary_type )
   call idd_set_char_qd_letters( char_index, ql, dl )
   if( len_trim(description) > 0 ) idd%chars(char_index)%description = trim(description)
   idd%chars(char_index)%flag = trim(flag)
end subroutine read_one_custom_characteristic

!
!#########################      T  R  A  C  E  S         ####################################
!

!
!---   s t a r t _ r e a d i n g _ o n e _ t r a c e  ( )   ----------------------------------
!
!
! starts reading one trace stored in upcoming sub-elements of the <Trace>
! all descriptive data are available in the attributes of the <Trace> element
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
subroutine start_reading_one_trace()

   character(len=IDD_MAX_TYPE_NAME)  :: trace_type
   character(len=IDD_MAX_LAYER_NAME) :: layer
   integer                           :: multiple
   character                         :: polarization
   integer                           :: num
   integer                           :: trace_number

   trace_type = VAL_STANDARD
   multiple = 1
   polarization = VAL_ORDINARY
   num = 0

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_TYPE)
         trace_type = trim(attribs(2,i))

      case (ATTR_LAYER)
         layer = trim(attribs(2,i))

      case (ATTR_MULTIPLE)
         read( attribs(2,i), *, IOSTAT=stat) multiple

      case (ATTR_POLARIZATION)
         polarization = attribs(2,i)(1:1)

      case (ATTR_NUM)
         read( attribs(2,i), *, IOSTAT=stat) num
      end select
   enddo

   if( num > 0 ) then
      trace_number = idd_get_total_traces() + 1
      idd%traces(trace_number)%available = .true.
      idd%traces(trace_number)%num = num
      idd%traces(trace_number)%trace_type = trace_type
      idd%traces(trace_number)%layer = layer
      idd%traces(trace_number)%multiple = multiple
      idd%traces(trace_number)%polarization = polarization
    endif
end subroutine start_reading_one_trace


!
!---   i d e n t i f y _ t r a c e _ v a l u e ()   ----------------------------------
!
!
! identifies which type of additional trace values (not frequency or range) are given 
! in the TraceValueList element
! unknown elements will be skipped
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     trace_value_type
!
subroutine identify_trace_value()

   trace_value_name = ''

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_NAME)
         trace_value_name = trim(attribs(2,i))

      case (ATTR_NO_VALUE)
         read( attribs(2,i), *, IOSTAT=stat) no_value_trace
      end select

   enddo
end subroutine identify_trace_value


subroutine read_trace_frequencies()
   integer :: trace_index
   trace_index = idd_get_total_traces() ! returns currently active (latest) trace
   call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                              data, no_data, idd%traces(trace_index)%frequencies, has_var )
end subroutine read_trace_frequencies

subroutine read_trace_ranges()
   integer :: trace_index
   trace_index = idd_get_total_traces() ! returns currently active (latest) trace
   call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                              data, no_data, idd%traces(trace_index)%ranges, has_var )
end subroutine read_trace_ranges

subroutine read_trace_values()
   integer :: trace_index
   trace_index = idd_get_total_traces() ! returns currently active (latest) trace

   select case ( trace_value_name )

   case (VAL_AMPLITUDE)
         idd%traces(trace_index)%amplitudes_available = .true.
         idd%traces(trace_index)%amplitude_no_value = no_value_trace
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                                     data, no_data, idd%traces(trace_index)%amplitudes, has_var )
   case (VAL_DOPPLER_SHIFT)
         idd%traces(trace_index)%doppler_shifts_available = .true.
         idd%traces(trace_index)%doppler_shift_no_value = no_value_trace
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                                     data, no_data, idd%traces(trace_index)%doppler_shifts, has_var )
   ! ToDo: implement reading of other types of trace value here

   end select

end subroutine read_trace_values


!
!#########################      P  R  O  F  I  L  E  S       #################################
!

!
!---   s t a r t _ r e a d i n g _ o n e _ p r o f i l e ( )   -------------------------------
!
!
! starts reading one profile stored in upcoming sub-elements of the <Profile>
! all descriptive data are available in the attributes of the <Profile> element
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
subroutine start_reading_one_profile()

   integer :: profile_number 
   profile_number = idd_get_total_profiles() + 1

   ! clear all attributes in case they are not provided
   idd%profiles(profile_number)%profile_type = VAL_VERTICAL
   idd%profiles(profile_number)%description = ''
   idd%profiles(profile_number)%algorithm = ''
   idd%profiles(profile_number)%algorithm_version = ''
   idd%profiles(profile_number)%available = .true.

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_TYPE)
         idd%profiles(profile_number)%profile_type = trim(attribs(2,i))

      case (ATTR_DESCRIPTION)
         idd%profiles(profile_number)%description = trim(attribs(2,i))

      case (ATTR_ALGORITHM)
         idd%profiles(profile_number)%algorithm = trim(attribs(2,i))

      case (ATTR_ALGORITHM_VERSION)
         idd%profiles(profile_number)%algorithm_version = trim(attribs(2,i))
      end select
   enddo

end subroutine start_reading_one_profile

!
!---   s t a r t _ r e a d i n g _ o n e _ t a b u l a t e d _ p r o f i l e ( )   -----------------------
!
!
! starts reading one tabulated profile stored in upcoming sub-elements of the <Tabulated>
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
subroutine start_reading_one_tabulated_profile()

   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_NUM)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%num_tabulated

      end select
   enddo

end subroutine start_reading_one_tabulated_profile

!
!---   r e a d _ p r o f i l e _ a l t i t u d e s ( )   ------------------------------------
!
!
subroutine read_profile_altitudes()
   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()
    call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                                data, no_data, idd%profiles(profile_index)%altitudes, has_var )
end subroutine read_profile_altitudes

!
!---   i d e n t i f y _ p r o f i l e _ v a l u e ()   -------------------------------------
!
!
! identifies which type of additional trace values (not frequency or range) are given 
! in the TraceValueList element
! unknown elements will be skipped
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     trace_value_type
!
subroutine identify_profile_value()

   profile_value_name = ''
   profile_bounds_available = .false.
   profile_lower_bounds_available = .false.
   profile_upper_bounds_available = .false.

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_NAME)
         profile_value_name = trim(attribs(2,i))

      case (ATTR_NO_VALUE)
         read( attribs(2,i), *, IOSTAT=stat) no_value_profile
      end select

   enddo
end subroutine identify_profile_value

!
!---   r e a d _ p r o f i l e _ v a l u e s ( )   ------------------------------------------
!
!
subroutine read_profile_values()
   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()
   select case ( profile_value_name )
      case (VAL_PLASMA_DENSITY)
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                                     data, no_data, idd%profiles(profile_index)%plasma_densities, has_var )
      case (VAL_PLASMA_FREQUENCY)
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                                   data, no_data, idd%profiles(profile_index)%plasma_frequencies, has_var )
      ! Implement reading of other profile values here

   end select

end subroutine read_profile_values

!
!---   r e a d _ p r o f i l e _ b o u n d s ( )   ------------------------------------------
!
!
subroutine read_profile_bounds()
   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()
   select case ( profile_value_name )
      case (VAL_PLASMA_DENSITY)
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                                   data, no_data, idd%profiles(profile_index)%plasma_density_bounds, has_var )
         idd%profiles(profile_index)%plasma_density_uncertainty_available = .true.
      
      case (VAL_PLASMA_FREQUENCY)
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, &
                                   data, no_data, idd%profiles(profile_index)%plasma_frequency_bounds, has_var )
         idd%profiles(profile_index)%plasma_frequency_uncertainty_available = .true.

      ! Implement reading of other profile value bounds here

   end select
   profile_bounds_available = .true.
end subroutine read_profile_bounds

!
!---   r e a d _ p r o f i l e _ l o w e r _ b o u n d s ( )   ------------------------------------------
!
!
subroutine read_profile_lower_bounds()
   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()
   select case ( profile_value_name )
      case (VAL_PLASMA_DENSITY)
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, data, no_data, &
                                   idd%profiles(profile_index)%plasma_density_lower_bounds, has_var )
         
         idd%profiles(profile_index)%plasma_density_uncertainty_available = .true.

      case (VAL_PLASMA_FREQUENCY)
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, data, no_data, &
                                   idd%profiles(profile_index)%plasma_frequency_lower_bounds, has_var )
         
         idd%profiles(profile_index)%plasma_frequency_uncertainty_available = .true.
      
      ! Implement reading of other profile value bounds here

   end select
   profile_lower_bounds_available = .true.
end subroutine read_profile_lower_bounds

!
!---   r e a d _ p r o f i l e _ u p p e r _ b o u n d s ( )   ------------------------------------------
!
!
subroutine read_profile_upper_bounds()
   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()
   select case ( profile_value_name )
      case (VAL_PLASMA_DENSITY)
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, data, no_data, &
                                   idd%profiles(profile_index)%plasma_density_upper_bounds, has_var )
         idd%profiles(profile_index)%plasma_density_uncertainty_available = .true.
      
      case (VAL_PLASMA_FREQUENCY)
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, data, no_data, &
                                   idd%profiles(profile_index)%plasma_frequency_upper_bounds, has_var )
         idd%profiles(profile_index)%plasma_frequency_uncertainty_available = .true.

      ! Implement reading of other profile value bounds here
      
   end select
   profile_upper_bounds_available = .true.

end subroutine read_profile_upper_bounds


!
!---   f i n a l i z e _ p r o f i l e _ v a l u e _ u n c e r t a i n t i e s ( )   ------------------------------------------
!
!
subroutine finalize_profile_value_uncertainties()
   ! current profile that is being read
   integer :: profile_index, total_altitudes, k
   profile_index = idd_get_total_profiles()
   total_altitudes = idd%profiles(profile_index)%num_tabulated
   select case ( profile_value_name )
      case (VAL_PLASMA_DENSITY)
         call fill_missing_bounds( &
            idd%profiles(profile_index)%plasma_densities, &
            idd%profiles(profile_index)%plasma_density_lower_bounds, &
            idd%profiles(profile_index)%plasma_density_upper_bounds, &
            idd%profiles(profile_index)%plasma_density_bounds, &
            total_altitudes )
      case (VAL_PLASMA_FREQUENCY)
         call fill_missing_bounds( &
            idd%profiles(profile_index)%plasma_frequencies, &
            idd%profiles(profile_index)%plasma_frequency_lower_bounds, &
            idd%profiles(profile_index)%plasma_frequency_upper_bounds, &
            idd%profiles(profile_index)%plasma_frequency_bounds, &
            total_altitudes )
      ! Implement reading of other profile values here
   end select

end subroutine finalize_profile_value_uncertainties

!
!---   f i l l _ m i s s i n g _ b o u n d s ( )   ------------------------------------------------
!
!
subroutine fill_missing_bounds( values, lower_bounds, upper_bounds, bounds, length )
   real(8), dimension(:), pointer         :: values, lower_bounds, upper_bounds, bounds
   integer                                :: length, k

   ! Use global state flags to understand which bounds are available

   ! Only upper bounds were provided, not lower. Assume symmetric bound
   if( profile_upper_bounds_available .and. .not.profile_lower_bounds_available ) then
      allocate( lower_bounds(1:length) )
      do k=1, length
         lower_bounds(k) = 2 * values(k) - upper_bounds(k)
      enddo
      profile_lower_bounds_available = .true.
   endif

   ! Only lower bounds were provided, not upper. Assume symmetric bound
   if( profile_lower_bounds_available .and. .not.profile_upper_bounds_available ) then
      allocate( upper_bounds(1:length) )
      do k=1, length
         upper_bounds(k) = 2 * values(k) - lower_bounds(k)
      enddo
      profile_upper_bounds_available  = .true.
   endif

   ! Only lower/upper bounds were provided, not symmetric. Derive pessimistic symmetric bound
   if( profile_upper_bounds_available .and. .not.profile_bounds_available ) then
      allocate( bounds(1:length) )
      do k=1, length
         if( lower_bounds(k) .eq. -1. ) then 
            bounds(k) = upper_bounds(k) - values(k)
         else
            bounds(k) = max( values(k) - lower_bounds(k), upper_bounds(k) - values(k) )
         endif
      enddo
      profile_bounds_available  = .true.
   endif

   ! Only symmetric bounds were provided, derive lower/upper bounds 
   if( profile_bounds_available .and. .not.profile_lower_bounds_available ) then
      allocate( lower_bounds(1:length) )
      allocate( upper_bounds(1:length) )
      do k=1, length
         lower_bounds(k) = values(k) - bounds(k)
         upper_bounds(k) = values(k) + bounds(k)
      enddo
      profile_lower_bounds_available  = .true.
      profile_upper_bounds_available  = .true.
   endif

end subroutine fill_missing_bounds

!
!---   f i n a l i z e _ p r o f i l e _ u n c e r t a i n t i e s ( )   --------------------------
!
!
subroutine finalize_profile_uncertainties()
   ! current profile that is being read
   integer :: profile_index, total_altitudes, k
   profile_index = idd_get_total_profiles()
   total_altitudes = idd%profiles(profile_index)%num_tabulated

   if(idd%profiles(profile_index)%plasma_frequency_uncertainty_available .and. &
      idd%profiles(profile_index)%plasma_density_uncertainty_available ) return

   if(.not. idd%profiles(profile_index)%plasma_frequency_uncertainty_available .and. &
      .not. idd%profiles(profile_index)%plasma_density_uncertainty_available ) return

   ! only one of the two uncertainties for density and frequency were given in the record

   if(idd%profiles(profile_index)%plasma_frequency_uncertainty_available ) then

      allocate( idd%profiles(profile_index)%plasma_density_bounds(1:total_altitudes) )
      allocate( idd%profiles(profile_index)%plasma_density_lower_bounds(1:total_altitudes) )
      allocate( idd%profiles(profile_index)%plasma_density_upper_bounds(1:total_altitudes) )
      do k=1, total_altitudes
         idd%profiles(profile_index)%plasma_density_bounds(k) = &
            idd_frequency_to_density_cm3( idd%profiles(profile_index)%plasma_frequency_bounds(k) )
         idd%profiles(profile_index)%plasma_density_lower_bounds(k) = &
            idd_frequency_to_density_cm3( idd%profiles(profile_index)%plasma_frequency_lower_bounds(k) )
         idd%profiles(profile_index)%plasma_density_upper_bounds(k) = &
            idd_frequency_to_density_cm3( idd%profiles(profile_index)%plasma_frequency_upper_bounds(k) )
      enddo
   endif

   if(idd%profiles(profile_index)%plasma_density_uncertainty_available ) then

      allocate( idd%profiles(profile_index)%plasma_frequency_bounds(1:total_altitudes) )
      allocate( idd%profiles(profile_index)%plasma_frequency_lower_bounds(1:total_altitudes) )
      allocate( idd%profiles(profile_index)%plasma_frequency_upper_bounds(1:total_altitudes) )
      do k=1, total_altitudes
         idd%profiles(profile_index)%plasma_frequency_bounds(k) = &
            idd_density_to_frequency( idd%profiles(profile_index)%plasma_density_bounds(k) )
         idd%profiles(profile_index)%plasma_frequency_lower_bounds(k) = &
            idd_density_to_frequency( idd%profiles(profile_index)%plasma_density_lower_bounds(k) )
         idd%profiles(profile_index)%plasma_frequency_upper_bounds(k) = &
            idd_density_to_frequency( idd%profiles(profile_index)%plasma_density_upper_bounds(k) )
      enddo
   endif
   idd%profiles(profile_index)%plasma_density_uncertainty_available = .true.
   idd%profiles(profile_index)%plasma_frequency_uncertainty_available = .true.

end subroutine finalize_profile_uncertainties

!
!---   r e a d _ p r o f i l e _ s h i f t e d _ c h e b y s h e v _ a t t r i b u t e s ( )   ----------
!
!
subroutine read_profile_shifted_chebyshev_attributes()

   tmp_cheb%region = ''
   tmp_cheb%num = 0
   tmp_cheb%z_half_Nm_km = NO_VALUE_CHAR

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_REGION)
         tmp_cheb%region = trim(attribs(2,i))

      case (ATTR_NUM)
         read( attribs(2,i), *, IOSTAT=stat) tmp_cheb%num

      case (ATTR_START_FREQUENCY)
         read( attribs(2,i), *, IOSTAT=stat) tmp_cheb%start_freq_MHz

      case (ATTR_END_FREQUENCY)
         read( attribs(2,i), *, IOSTAT=stat) tmp_cheb%end_freq_MHz

      case (ATTR_PEAK_HEIGHT)
         read( attribs(2,i), *, IOSTAT=stat) tmp_cheb%peak_height_km

      case (ATTR_ERROR)
         read( attribs(2,i), *, IOSTAT=stat) tmp_cheb%fit_error

      case (ATTR_Z_HALF_NM)
         read( attribs(2,i), *, IOSTAT=stat) tmp_cheb%z_half_Nm_km

      end select

   enddo

end subroutine read_profile_shifted_chebyshev_attributes

!
!---   r e a d _ p r o f i l e _ s h i f t e d _ c h e b y s h e v _ c o e f s ( )   ----------
!
!
subroutine read_profile_shifted_chebyshev_coefs()

   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()

   select case ( tmp_cheb%region )
      case (VAL_LAYER_F2)
         idd%profiles(profile_index)%f2_chebyshev = tmp_cheb
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, data, &
                                   no_data, idd%profiles(profile_index)%f2_chebyshev%c, has_var )
         if( tmp_cheb%num .eq. 0 ) idd%profiles(profile_index)%f2_chebyshev%num = size( idd%profiles(profile_index)%f2_chebyshev%c )
      case (VAL_LAYER_F1)
         idd%profiles(profile_index)%f1_chebyshev = tmp_cheb
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, data, &
                                   no_data, idd%profiles(profile_index)%f1_chebyshev%c, has_var )
         if( tmp_cheb%num .eq. 0 ) idd%profiles(profile_index)%f1_chebyshev%num = size( idd%profiles(profile_index)%f1_chebyshev%c )
      case (VAL_LAYER_E)
         idd%profiles(profile_index)%e_chebyshev = tmp_cheb
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, data, &
                                   no_data, idd%profiles(profile_index)%e_chebyshev%c, has_var )
         if( tmp_cheb%num .eq. 0 ) idd%profiles(profile_index)%e_chebyshev%num = size( idd%profiles(profile_index)%e_chebyshev%c )
      case (VAL_LAYER_F)
         idd%profiles(profile_index)%f2_chebyshev = tmp_cheb
         call read_xml_real_array( info, tag, endtag, attribs, no_attribs, data, &
                                   no_data, idd%profiles(profile_index)%f2_chebyshev%c, has_var )
         if( tmp_cheb%num .eq. 0 ) idd%profiles(profile_index)%f2_chebyshev%num = size( idd%profiles(profile_index)%f2_chebyshev%c )
   end select

end subroutine read_profile_shifted_chebyshev_coefs

!
!---   s t a r t _ r e a d i n g _ q p _ s e g m e n t s ( )   -----------------------
!
!
! starts reading qp segments stored as sub-elements of the <QuasiParabolicList>
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
subroutine start_reading_qp_segments()

   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()

   idd%profiles(profile_index)%num_qp_segments = 0

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_NUM)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%num_qp_segments

      case (ATTR_EARTH_RADIUS)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%earth_radius

      end select
   enddo

   ! allocate memory for segments
   ! deallocation is done in idd_reset method
   if( idd%profiles(profile_index)%num_qp_segments > 0 ) &
       allocate( idd%profiles(profile_index)%qps( idd%profiles(profile_index)%num_qp_segments ) )

end subroutine start_reading_qp_segments

!
!---   r e a d _ o n e _ q p _ s e g m e n t ( )   -----------------------
!
!
! read one qp segment specified in <QuasiParabolic> element attributes
!
! Input global variables: 
!     attribs (array of attributes) 
!     no_attribs (number of attributes)
!
! Output global variables
!     idd   
subroutine read_one_qp_segment()

   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()
   
   ! reset id to make sure the <QuasiParabolic> element has a good ID value
   tmp_qp%id = 0

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_ID)
         read( attribs(2,i), *, IOSTAT=stat) tmp_qp%id

      case (ATTR_START_DISTANCE)
         read( attribs(2,i), *, IOSTAT=stat) tmp_qp%start_distance_km

      case (ATTR_END_DISTANCE)
         read( attribs(2,i), *, IOSTAT=stat) tmp_qp%end_distance_km

      case (ATTR_A)
         read( attribs(2,i), *, IOSTAT=stat) tmp_qp%a

      case (ATTR_B)
         read( attribs(2,i), *, IOSTAT=stat) tmp_qp%b

      case (ATTR_C)
         read( attribs(2,i), *, IOSTAT=stat) tmp_qp%c

      case (ATTR_ERROR)
         read( attribs(2,i), *, IOSTAT=stat) tmp_qp%fit_error

      end select
   enddo

   if( tmp_qp%id .ge. 0 .and. tmp_qp%id < idd%profiles(profile_index)%num_qp_segments ) &
       idd%profiles(profile_index)%qps( tmp_qp%id+1 ) = tmp_qp

end subroutine read_one_qp_segment

!
!---   r e a d _ p r o f i l e _ v a l l e y ( )   --------------------------------------------------
!
!
subroutine read_profile_valley()

   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_MODEL)
         idd%profiles(profile_index)%valley_model = trim(attribs(2,i))

      case (ATTR_WIDTH)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%valley_width_km

      case (ATTR_DEPTH)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%valley_depth_pfu

      case (ATTR_START_HEIGHT)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%valley_start_height_km

      case (ATTR_START_FREQUENCY)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%valley_start_freq_MHz

      end select
   enddo

end subroutine read_profile_valley

!
!---   r e a d _ p r o f i l e _ t o p s i d e _ C h a p m a n ( )   -------------------------------------
!
!
subroutine read_profile_topside_chapman_layer()

   ! current profile that is being read
   integer :: profile_index 
   profile_index = idd_get_total_profiles()

   idd%profiles(profile_index)%top_chap%peak_height_km = 0
   idd%profiles(profile_index)%top_chap%peak_density_cm3 = 0
   idd%profiles(profile_index)%top_chap%peak_scale_height_km = 0

   do i=1,no_attribs

      select case ( trim(attribs(1,i)) )

      case (ATTR_PEAK_HEIGHT)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%top_chap%peak_height_km

      case (ATTR_PEAK_DENSITY)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%top_chap%peak_density_cm3

      case (ATTR_PEAK_SCALE_HEIGHT)
         read( attribs(2,i), *, IOSTAT=stat) idd%profiles(profile_index)%top_chap%peak_scale_height_km

      end select
   enddo
   if( idd%profiles(profile_index)%top_chap%peak_height_km > 0 ) &
     idd%profiles(profile_index)%top_chap%available = .true.


end subroutine read_profile_topside_chapman_layer


!###################################################################################################
!
!
!      S   A   O   X   M   L       W   R   I   T   I   N   G       F   U   N   C   T   I   O   N   S
!
!
!###################################################################################################

!
!---   s a o x m l _ o p e n _ o u t p u t _ f i l e  ( )    ----------------------------------------
!
subroutine saoxml_open_output_file( fname )

    character(len=*) :: fname

! Assign the xml-filename to fname and open the file
   mustread = .false.
   call xml_open(info,fname,mustread)
! Check for errors
   if (xml_error(info)) then
   ! ToDo: Handle the errors
   else
      call xml_options(info,ignore_whitespace = .true.)
      call saoxml_write_open_tag( TAG_SAO_RECORD_LIST )
   endif
end subroutine saoxml_open_output_file

!
!---   s a o x m l _ c l o s e _ o u t p u t _ f i l e  ( )       ---------------------------------
!
subroutine saoxml_close_output_file()
   call saoxml_write_close_tag( TAG_SAO_RECORD_LIST )
   call xml_close(info)
   return
end subroutine saoxml_close_output_file


subroutine saoxml_write_open_tag( tag )
   character(len=*) :: tag
   write( info%lun, '(3a)' ) '<', tag, '>'
end subroutine saoxml_write_open_tag

subroutine saoxml_write_close_tag( tag )
   character(len=*) :: tag
   write( info%lun, '(3a)' ) '</', tag, '>'
end subroutine saoxml_write_close_tag

subroutine saoxml_write_attribute_line( attribute, string_value )
   character(len=*) :: attribute, string_value
   write( info%lun, '(4a)' ) attribute, '="', string_value,'"'
end subroutine saoxml_write_attribute_line

!
!---   s a o x m l _ w r i t e _ o n e _ r e c o r d ( )      -------------------------------------
!
subroutine saoxml_write_one_record()

   call saoxml_write_sao_record_opening()
   call saoxml_write_system_info()
   call saoxml_write_characteristics()
   call saoxml_write_traces()
   call saoxml_write_profiles()
   call saoxml_write_close_tag( TAG_SAO_RECORD )

end subroutine saoxml_write_one_record


!
!###############      T  O  P    L  E  V  E  L    M  E  T  A  D  A  T  A          #################
!

!
!---   w r i t e _ s a o _ r e c o r d _ o p e n i n g ( )      -----------------------------------
!
subroutine saoxml_write_sao_record_opening()

   write( info%lun, '(3a)' ) '<', TAG_SAO_RECORD
   ! format version
   write( buffer, '(i0,a,i0)') MAJOR_VERSION_OUTPUT, '.',MINOR_VERSION_OUTPUT
   call saoxml_write_attribute_line( ATTR_FORMAT_VERSION, trim(buffer) )
   ! start time
   write( buffer, '(i4.4,a,i2.2,a,i2.2,1x,a,i3.3,1x,i2.2,a,i2.2,a,i2.2,a,i3.3)') &
      idd%time%year,'-',idd%time%month,'-',idd%time%day_of_month,'-',&
      idd%time%day_of_year,idd%time%hour,':',idd%time%minute,':',&
      idd%time%second,'.',idd%time%millisecond
   call saoxml_write_attribute_line( ATTR_START_TIME_UTC, trim(buffer) )
   ! URSI code
   call saoxml_write_attribute_line( ATTR_URSI_CODE, idd%location%ursi_code )
   ! Station name
   call saoxml_write_attribute_line( ATTR_STATION_NAME, trim(idd%location%station_name) )
   ! Geolatitude
   call saoxml_write_attribute_line( ATTR_GEO_LATITUDE, trim(compact(idd%location%geo_latitude)) )
   ! Geolongitude
   call saoxml_write_attribute_line( ATTR_GEO_LONGITUDE, trim(compact(idd%location%geo_longitude) ) )
   ! Source
   call saoxml_write_attribute_line( ATTR_SOURCE, trim(idd%system%source) )
   ! Source type
   call saoxml_write_attribute_line( ATTR_SOURCE_TYPE, trim(idd%system%source_type) )
   ! Scaler type
   if(idd%system%autoscaled) then
      call saoxml_write_attribute_line( ATTR_SCALER_TYPE, VAL_AUTO )
   else
      call saoxml_write_attribute_line( ATTR_SCALER_TYPE, VAL_MANUAL )
   endif
   write( info%lun, '(a)' ) '>'
end subroutine saoxml_write_sao_record_opening

!
!---   w r i t e _ s y s t e m _ i n f o _ o p e n i n g ( )      -----------------------------------
!
subroutine saoxml_write_system_info()

   ! write the opening tag with one possible attribute (so far)
   if( idd%location%uml_sid .eq. 0 ) then 
      call saoxml_write_open_tag( TAG_SYSTEM_INFO )
   else
      write( info%lun, '(5a,I3.3,a)' ) '<',TAG_SYSTEM_INFO, ' ',ATTR_UML_STATION_ID,'="',idd%location%uml_sid,'">'
   endif
   ! write frequency stepping info
   no_attribs = 2
   attribs(1,1) = ATTR_START_FREQUENCY
   attribs(2,1) = compact( idd%system%ionogram%start_frequency_MHz ) 
   attribs(1,2) = ATTR_STOP_FREQUENCY
   attribs(2,2) = compact( idd%system%ionogram%stop_frequency_MHz )
   no_data = 0
   call xml_put(info, TAG_FREQUENCY_STEPPING, attribs, no_attribs, data, no_data, 'open')

   ! write frequency stepping details
   no_attribs = 2
   attribs(1,1) = ATTR_STEP
   attribs(2,1) = compact( idd%system%ionogram%frequency_step )
   attribs(1,2) = ATTR_UNITS
   attribs(2,2) = idd%system%ionogram%frequency_step_units
   no_data = 0
   if( idd%system%ionogram%frequency_stepping_rule .eq. IDD_LINEAR_STEPPING ) then
      call xml_put(info, TAG_LINEAR_STEPPING, attribs, no_attribs, data, no_data, 'elem')
   endif
   if( idd%system%ionogram%frequency_stepping_rule .eq. IDD_LOG_STEPPING ) then
      call xml_put(info, TAG_LOG_STEPPING, attribs, no_attribs, data, no_data, 'elem')
   endif
   call xml_put(info, TAG_FREQUENCY_STEPPING, attribs, no_attribs, data, no_data, 'close')
   
   ! write restricted frequency bands
   if( idd%system%ionogram%total_restricted_freq_bands > 0 ) then
      no_attribs = 2
      attribs(1,1) = ATTR_NUM
      attribs(2,1) = icompact( idd%system%ionogram%total_restricted_freq_bands )
      attribs(1,2) = ATTR_UNITS
      attribs(2,2) = idd%system%ionogram%restricted_freq_units
      no_data = 0
      call xml_put(info, TAG_RESTRICTED_FREQUENCY_LIST, attribs, no_attribs, data, no_data, 'open')
      ! lower limits
      ! use compact string form (optimal precision unknown because of unknown units)
      no_attribs = 0
      call saoxml_copy_array_to_data1( idd%system%ionogram%lower_restricted_freq_limits, &
                                idd%system%ionogram%total_restricted_freq_bands)
      call xml_put(info, TAG_LOWER_LIMIT_LIST, attribs, no_attribs, data, no_data, 'elem')
      ! upper limits
      call saoxml_copy_array_to_data1( idd%system%ionogram%upper_restricted_freq_limits, &
                                idd%system%ionogram%total_restricted_freq_bands)
      call xml_put(info, TAG_UPPER_LIMIT_LIST, attribs, no_attribs, data, no_data, 'elem')
      no_data = 0
      call xml_put(info, TAG_RESTRICTED_FREQUENCY_LIST, attribs, no_attribs, data, no_data, 'close')
   endif

   ! write range stepping info
   no_attribs = 2
   attribs(1,1) = ATTR_START_RANGE
   attribs(2,1) = compact( idd%system%ionogram%start_range_km ) 
   attribs(1,2) = ATTR_STOP_RANGE
   attribs(2,2) = compact( idd%system%ionogram%stop_range_km )
   no_data = 0
   call xml_put(info, TAG_RANGE_STEPPING, attribs, no_attribs, data, no_data, 'open')

   ! write range stepping details
   no_attribs = 2
   attribs(1,1) = ATTR_STEP
   attribs(2,1) = compact( idd%system%ionogram%range_step )
   attribs(1,2) = ATTR_UNITS
   attribs(2,2) = idd%system%ionogram%range_step_units
   no_data = 0
   if( idd%system%ionogram%range_stepping_rule .eq. IDD_LINEAR_STEPPING ) then
      call xml_put(info, TAG_LINEAR_STEPPING, attribs, no_attribs, data, no_data, 'elem')
   endif
   if( idd%system%ionogram%range_stepping_rule .eq. IDD_LOG_STEPPING ) then
      call xml_put(info, TAG_LOG_STEPPING, attribs, no_attribs, data, no_data, 'elem')
   endif
   call xml_put(info, TAG_RANGE_STEPPING, attribs, no_attribs, data, no_data, 'close')

   ! write autoscaler software details
   if( idd%system%autoscaled ) then
      no_attribs = 3
      attribs(1,1) = ATTR_NAME
      attribs(2,1) = idd%system%auto_scaler
      attribs(1,2) = ATTR_VERSION
      attribs(2,2) = idd%system%auto_scaler_version
      attribs(1,3) = ATTR_ARTIST_FLAGS
      write( buffer, '(20(I0,1x))') idd%system%artist_flags
      attribs(2,3) = buffer
      no_data = 0
      call xml_put(info, TAG_AUTO_SCALER, attribs, no_attribs, data, no_data, 'elem')
   endif

   ! write comments
   if( len_trim( idd%system%comments ) > 0) then
      no_attribs = 0
      no_data = 1
      data(1) = idd%system%comments
      call xml_put(info, TAG_COMMENTS, attribs, no_attribs, data, no_data, 'elem')
   endif

   ! contact person
   if( idd%system%contact%available .and. len_trim( idd%system%contact%name ) > 0 ) then

      no_attribs = 0
      no_data = 0
      call xml_put(info, TAG_CONTACT_PERSON, attribs, no_attribs, data, no_data, 'open')
      no_data = 1
      data(1) = idd%system%contact%name
      call xml_put(info, TAG_NAME, attribs, no_attribs, data, no_data, 'elem')
      data(1) = idd%system%contact%affiliation
      call xml_put(info, TAG_AFFILIATION, attribs, no_attribs, data, no_data, 'elem')
      data(1) = idd%system%contact%address
      call xml_put(info, TAG_ADDRESS, attribs, no_attribs, data, no_data, 'elem')
      data(1) = idd%system%contact%email
      call xml_put(info, TAG_EMAIL, attribs, no_attribs, data, no_data, 'elem')
      call xml_put(info, TAG_CONTACT_PERSON, attribs, no_attribs, data, no_data, 'close')
   endif

   ! write solar-terrestrial data
   no_attribs = 0
   no_data = 0
   call xml_put(info, TAG_SOLAR_TERRESTRIAL_DATA, attribs, no_attribs, data, no_data, 'open')
   ! write gyrofrequency
   if( idd%geospace%gyro_frequency_MHz .ne. NO_VALUE_CHAR ) then
      no_attribs = 3
      attribs(1,1) = ATTR_VAL
      attribs(2,1) = compact( idd%geospace%gyro_frequency_MHz ) 
      attribs(1,2) = ATTR_MODEL
      attribs(2,2) = idd%geospace%magfield_model_name
      attribs(1,3) = ATTR_ALTITUDE
      attribs(2,3) = compact( idd%geospace%magfield_model_altitude ) 
      no_data = 0
      call xml_put(info, TAG_GYROFREQUENCY, attribs, no_attribs, data, no_data, 'elem')
   endif
   ! write dip angle
   if( idd%geospace%dip_angle_deg .ne. NO_VALUE_CHAR ) then
      no_attribs = 3
      attribs(1,1) = ATTR_VAL
      attribs(2,1) = compact( idd%geospace%dip_angle_deg ) 
      attribs(1,2) = ATTR_MODEL
      attribs(2,2) = idd%geospace%magfield_model_name
      attribs(1,3) = ATTR_ALTITUDE
      attribs(2,3) = compact( idd%geospace%magfield_model_altitude ) 
      no_data = 0
      call xml_put(info, TAG_DIP_ANGLE, attribs, no_attribs, data, no_data, 'elem')
   endif
   ! write sunspot number
   if( idd%geospace%sunspot_number .ne. NO_VALUE_CHAR ) then
      no_attribs = 2
      attribs(1,1) = ATTR_VAL
      attribs(2,1) = compact( idd%geospace%sunspot_number ) 
      attribs(1,2) = ATTR_STATUS
      attribs(2,2) = idd%geospace%sunspot_number_status
      no_data = 0
      call xml_put(info, TAG_SUNSPOT_NUMBER, attribs, no_attribs, data, no_data, 'elem')
   endif
   call xml_put(info, TAG_SOLAR_TERRESTRIAL_DATA, attribs, no_attribs, data, no_data, 'close')

   ! digisonde preface
   if( len_trim( idd%system%digisonde_preface ) > 0 ) then
      no_attribs = 1
      attribs(1,1) = ATTR_VERSION
      attribs(2,1) = idd%system%digisonde_preface(1:2)
      no_data = 1
      data(1) = trim( idd%system%digisonde_preface )
      call xml_put(info, TAG_DIGISONDE_PREFACE, attribs, no_attribs, data, no_data, 'elem')
   endif

   call xml_put(info, TAG_SYSTEM_INFO, attribs, no_attribs, data, no_data, 'close')
end subroutine saoxml_write_system_info

!
!#################      C  H  A  R  A  C  T  E  R  I  S  T  I  C  S         #####################
!

!
!---   w r i t e _ c h a r a c t e r i s t i c s ( )    ---------------------
!
!
subroutine saoxml_write_characteristics()
   
   integer :: i, total_count, list_index
   total_count = 0
   do i=1, IDD_TOTAL_CHARS
      if(idd%chars(i)%value .ne. NO_VALUE_CHAR ) total_count = total_count+1
   enddo
   if( total_count .eq. 0 ) return

   ! opening tag with total number
   no_attribs = 1
   attribs(1,1) = ATTR_NUM
   attribs(2,1) = icompact( total_count )
   no_data = 0
   call xml_put(info, TAG_CHARACTERISTIC_LIST, attribs, no_attribs, data, no_data, 'open')
   
   ! write standard chars first
   no_data=0
   do i=1, IDD_TOTAL_STANDARD_CHARS
      if(idd%chars(i)%value .ne. NO_VALUE_CHAR ) then
      
         ! required items
         attribs(1,1) = ATTR_ID
         attribs(2,1) = URSI_CODE_CHAR(i)
         attribs(1,2) = ATTR_VAL
         attribs(2,2) = compact( idd%chars(i)%value )
         no_attribs = 2
         ! optional items
         no_attribs = no_attribs + 1
         attribs(1,no_attribs) = ATTR_NAME
         attribs(2,no_attribs) = trim(URSI_NAMES_CHAR(i))
         ! do not output Q an D for an autoscaled record, those must be "//" for all chars
         ! when records are edited, some values can still be autoscaled
         if( .not. idd%system%autoscaled ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_QL
            attribs(2,no_attribs) = idd%chars(i)%q
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_DL
            attribs(2,no_attribs) = idd%chars(i)%d
         endif

         ! bounds are possible
         if( idd%chars(i)%bound .ne. NO_VALUE_CHAR ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_BOUND
            attribs(2,no_attribs) = compact( idd%chars(i)%bound )
         endif
         if( idd%chars(i)%lower_bound .ne. NO_VALUE_CHAR ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_LOWER_BOUND
            attribs(2,no_attribs) = compact( idd%chars(i)%lower_bound )
         endif
         if( idd%chars(i)%upper_bound .ne. NO_VALUE_CHAR ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_UPPER_BOUND
            attribs(2,no_attribs) = compact( idd%chars(i)%upper_bound )
         endif
         if( len_trim(idd%chars(i)%boundary_type) > 0) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_BOUNDARY_TYPE
            attribs(2,no_attribs) = trim( idd%chars(i)%boundary_type )
         endif
         if( len_trim(idd%chars(i)%flag) > 0 ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_FLAG
            attribs(2,no_attribs) = trim(idd%chars(i)%flag)
         endif
         call xml_put(info, TAG_URSI, attribs, no_attribs, data, no_data, 'elem')

      endif
   enddo
  
  ! write modeled chars second
   no_data=0
   do i=1, IDD_TOTAL_MODELED_CHARS
      list_index = i+IDD_TOTAL_STANDARD_CHARS
      if(idd%chars(list_index)%value .ne. NO_VALUE_CHAR ) then
      
         ! required items
         attribs(1,1) = ATTR_NAME
         attribs(2,1) = trim(MODELED_CHAR_NAME(i))
         attribs(1,2) = ATTR_VAL
         attribs(2,2) = compact( idd%chars(list_index)%value )
         attribs(1,3) = ATTR_UNITS
         attribs(2,3) = trim( idd%chars(list_index)%units )
         no_attribs = 3
         ! optional items
         if( len_trim(idd%system%model_name(i)) > 0 ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_MODEL_NAME
            attribs(2,no_attribs) = trim(idd%system%model_name(i))
         endif
         if( len_trim(idd%system%model_options(i)) > 0 ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_MODEL_OPTIONS
            attribs(2,no_attribs) = trim(idd%system%model_options(i))
         endif
         call xml_put(info, TAG_MODELED, attribs, no_attribs, data, no_data, 'elem')
      endif
   enddo
 
  ! write custom chars last
   no_data=0
   do i=1, IDD_TOTAL_CUSTOM_CHARS
      list_index = i+IDD_TOTAL_STANDARD_CHARS+IDD_TOTAL_MODELED_CHARS
      if(idd%chars(list_index)%value .ne. NO_VALUE_CHAR ) then
         ! required items
         attribs(1,1) = ATTR_NAME
         attribs(2,1) = trim(CUSTOM_CHAR_NAME(i))
         attribs(1,2) = ATTR_VAL
         attribs(2,2) = compact( idd%chars(list_index)%value )
         attribs(1,3) = ATTR_UNITS
         attribs(2,3) = idd%chars(list_index)%units
         no_attribs = 3
         ! optional items
         ! description
         if( len_trim(idd%chars(list_index)%description) > 0 ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_DESCRIPTION
            attribs(2,no_attribs) = trim(idd%chars(list_index)%description)
         endif
         ! bounds are possible
         if( idd%chars(list_index)%bound .ne. NO_VALUE_CHAR ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_BOUND
            attribs(2,no_attribs) = compact( idd%chars(list_index)%bound )
         endif
         if( idd%chars(list_index)%lower_bound .ne. NO_VALUE_CHAR ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_LOWER_BOUND
            attribs(2,no_attribs) = compact( idd%chars(list_index)%lower_bound )
         endif
         if( len_trim(idd%chars(i)%boundary_type) > 0) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_BOUNDARY_TYPE
            attribs(2,no_attribs) = trim( idd%chars(i)%boundary_type )
         endif
         if( idd%chars(list_index)%upper_bound .ne. NO_VALUE_CHAR ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_UPPER_BOUND
            attribs(2,no_attribs) = compact( idd%chars(list_index)%upper_bound )
         endif
         ! do not output Q an D for an autoscaled record, those must be "//" for all chars
         ! when records are edited, some values can still be autoscaled
         if( .not. idd%system%autoscaled ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_QL
            attribs(2,no_attribs) = idd%chars(list_index)%q
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_DL
            attribs(2,no_attribs) = idd%chars(list_index)%d
         endif
         ! edited or validated flag is possible
         if( len_trim(idd%chars(i)%flag) >0 ) then
            no_attribs = no_attribs + 1
            attribs(1,no_attribs) = ATTR_FLAG
            attribs(2,no_attribs) = trim(idd%chars(i)%flag)
         endif
         call xml_put(info, TAG_CUSTOM, attribs, no_attribs, data, no_data, 'elem')
      endif
   enddo
   call xml_put(info, TAG_CHARACTERISTIC_LIST, attribs, no_attribs, data, no_data, 'close')
end subroutine saoxml_write_characteristics

!
!#########################      T  R  A  C  E  S         ####################################
!

!
!---   w r i t e _ t r a c e s ( )   --------------------------------------------------------------
!
!
subroutine saoxml_write_traces()

   integer:: itrace

   if( idd_get_total_traces() .eq. 0 ) return
   
   ! opening <TraceList> tag with total number
   no_attribs = 1
   attribs(1,1) = ATTR_NUM
   attribs(2,1) = icompact( idd_get_total_traces() )
   no_data = 0
   call xml_put(info, TAG_TRACE_LIST, attribs, no_attribs, data, no_data, 'open')
   do itrace=1, idd_get_total_traces()
      call saoxml_write_one_trace( itrace )
   enddo
   call xml_put(info, TAG_TRACE_LIST, attribs, no_attribs, data, no_data, 'close')

end subroutine saoxml_write_traces

subroutine saoxml_write_one_trace( trace_index )
   integer :: trace_index

   ! opening <Trace> tag with required attributes
   no_attribs = 4
   attribs(1,1) = ATTR_TYPE
   attribs(2,1) = trim(idd%traces(trace_index)%trace_type)
   attribs(1,2) = ATTR_LAYER
   attribs(2,2) = trim(idd%traces(trace_index)%layer)
   attribs(1,3) = ATTR_POLARIZATION
   attribs(2,3) = idd%traces(trace_index)%polarization
   attribs(1,4) = ATTR_NUM
   attribs(2,4) = icompact(idd%traces(trace_index)%num)
   ! optional attribute
   if( idd%traces(trace_index)%multiple > 1 ) then
      no_attribs = 5
      attribs(1,5) = ATTR_MULTIPLE
      attribs(2,5) = icompact(idd%traces(trace_index)%multiple)
   endif
   no_data = 0
   call xml_put(info, TAG_TRACE, attribs, no_attribs, data, no_data, 'open')
   ! store frequency list first
   no_attribs = 3
   attribs(1,1) = ATTR_TYPE
   attribs(2,1) = VAL_FLOAT
   attribs(1,2) = ATTR_UNITS
   attribs(2,2) = 'MHz'
   attribs(1,3) = ATTR_DESCRIPTION
   attribs(2,3) = 'Nominal Frequency'
   call saoxml_copy_array_to_data2( idd%traces(trace_index)%frequencies, idd%traces(trace_index)%num )
   call xml_put(info, TAG_FREQUENCY_LIST, attribs, no_attribs, data, no_data, 'elem')
   
   ! store range list second
   no_attribs = 3
   attribs(1,1) = ATTR_TYPE
   attribs(2,1) = VAL_FLOAT
   attribs(1,2) = ATTR_UNITS
   attribs(2,2) = 'km'
   attribs(1,3) = ATTR_DESCRIPTION
   attribs(2,3) = 'Group Range'
   call saoxml_copy_array_to_data2( idd%traces(trace_index)%ranges, idd%traces(trace_index)%num )
   call xml_put(info, TAG_RANGE_LIST, attribs, no_attribs, data, no_data, 'elem')

   ! store optional trace values
   if( idd%traces(trace_index)%amplitudes_available ) then
      no_attribs = 5
      attribs(1,1) = ATTR_NAME
      attribs(2,1) = VAL_AMPLITUDE
      attribs(1,2) = ATTR_TYPE
      attribs(2,2) = VAL_INTEGER
      attribs(1,3) = ATTR_UNITS
      attribs(2,3) = 'dB'
      attribs(1,4) = ATTR_NO_VALUE
      attribs(2,4) = compact( idd%traces(trace_index)%amplitude_no_value )
      attribs(1,5) = ATTR_DESCRIPTION
      attribs(2,5) = 'Relative Amplitude'
      call saoxml_copy_array_to_data2( idd%traces(trace_index)%amplitudes, idd%traces(trace_index)%num )
      call xml_put(info, TAG_TRACE_VALUE_LIST, attribs, no_attribs, data, no_data, 'elem')

   endif
   if( idd%traces(trace_index)%doppler_shifts_available ) then
      no_attribs = 5
      attribs(1,1) = ATTR_NAME
      attribs(2,1) = VAL_DOPPLER_SHIFT
      attribs(1,2) = ATTR_TYPE
      attribs(2,2) = VAL_FLOAT
      attribs(1,3) = ATTR_UNITS
      attribs(2,3) = 'Hz'
      attribs(1,4) = ATTR_NO_VALUE
      attribs(2,4) = compact( idd%traces(trace_index)%doppler_shift_no_value )
      attribs(1,5) = ATTR_DESCRIPTION
      attribs(2,5) = 'Doppler Frequency Shift'
      call saoxml_copy_array_to_data2( idd%traces(trace_index)%doppler_shifts, idd%traces(trace_index)%num )
      call xml_put(info, TAG_TRACE_VALUE_LIST, attribs, no_attribs, data, no_data, 'elem')
   endif
   call xml_put(info, TAG_TRACE, attribs, no_attribs, data, no_data, 'close')

end subroutine saoxml_write_one_trace

!
!#########################      P  R  O  F  I  L  E  S       #################################
!

!
!---   w r i t e _ p r o f i l e s ( )   -----------------------------------------------------
!
!
subroutine saoxml_write_profiles()

   integer:: iprofile

   if( idd_get_total_profiles() .eq. 0 ) return
   
   ! opening <TraceList> tag with total number
   no_attribs = 1
   attribs(1,1) = ATTR_NUM
   attribs(2,1) = icompact( idd_get_total_profiles() )
   no_data = 0
   call xml_put(info, TAG_PROFILE_LIST, attribs, no_attribs, data, no_data, 'open')
   do iprofile=1, idd_get_total_profiles()
      call saoxml_write_one_profile( iprofile )
   enddo
   call xml_put(info, TAG_PROFILE_LIST, attribs, no_attribs, data, no_data, 'close')
end subroutine saoxml_write_profiles

!
!---   w r i t e _ o n e _ p r o f i l e ( )   ------------------------------------------------
!
!
subroutine saoxml_write_one_profile( profile_index )
   integer :: profile_index, k

   ! opening <Profile> tag with required attributes
   no_attribs = 2
   attribs(1,1) = ATTR_ALGORITHM
   attribs(2,1) = trim(idd%profiles(profile_index)%algorithm)
   attribs(1,2) = ATTR_ALGORITHM_VERSION
   attribs(2,2) = trim( idd%profiles(profile_index)%algorithm_version)
   ! optional attributes
   if( len_trim( idd%profiles(profile_index)%profile_type ) > 0 ) then
      no_attribs = no_attribs + 1
      attribs(1,no_attribs ) = ATTR_TYPE
      attribs(2,no_attribs ) = trim(idd%profiles(profile_index)%profile_type)
   endif
   if( len_trim( idd%profiles(profile_index)%description ) > 0 ) then
      no_attribs = no_attribs + 1
      attribs(1,no_attribs ) = ATTR_DESCRIPTION
      attribs(2,no_attribs ) = trim(idd%profiles(profile_index)%description)
   endif
   no_data = 0
   call xml_put(info, TAG_PROFILE, attribs, no_attribs, data, no_data, 'open')
   
   ! store tabulated representation first
   if( idd%profiles(profile_index)%num_tabulated > 0 ) then
      no_attribs = 1
      no_data = 0
      attribs(1,1) = ATTR_NUM
      attribs(2,1) = icompact( idd%profiles(profile_index)%num_tabulated )
      call xml_put(info, TAG_TABULATED, attribs, no_attribs, data, no_data, 'open')
      ! altitudes
      no_attribs = 3
      attribs(1,1) = ATTR_TYPE
      attribs(2,1) = VAL_FLOAT
      attribs(1,2) = ATTR_UNITS
      attribs(2,2) = 'km'
      attribs(1,3) = ATTR_DESCRIPTION
      attribs(2,3) = 'Altitude above station level'
      call saoxml_copy_array_to_data2( idd%profiles(profile_index)%altitudes, &
                                       idd%profiles(profile_index)%num_tabulated )
      call xml_put(info, TAG_ALTITUDE_LIST, attribs, no_attribs, data, no_data, 'elem')
   
      ! plasma frequencies
      no_attribs = 3
      attribs(1,1) = ATTR_NAME
      attribs(2,1) = VAL_PLASMA_FREQUENCY
      attribs(1,2) = ATTR_TYPE
      attribs(2,2) = VAL_FLOAT
      attribs(1,3) = ATTR_UNITS
      attribs(2,3) = 'MHz'
      call saoxml_copy_array_to_data2( idd%profiles(profile_index)%plasma_frequencies, &
                                       idd%profiles(profile_index)%num_tabulated )
      call xml_put(info, TAG_PROFILE_VALUE_LIST, attribs, no_attribs, data, no_data, 'elem')

      if( idd%profiles(profile_index)%plasma_density_uncertainty_available ) then

         ! uncertainty bounds
         no_attribs = 0
         info%level = 0 ! to avoid indenting
         call saoxml_copy_array_to_data2( idd%profiles(profile_index)%plasma_frequency_bounds, &
                                          idd%profiles(profile_index)%num_tabulated )
         call xml_put(info, TAG_BOUND_LIST, attribs, no_attribs, data, no_data, 'elem')
         call saoxml_copy_array_to_data2( idd%profiles(profile_index)%plasma_frequency_lower_bounds, &
                                          idd%profiles(profile_index)%num_tabulated )
         call xml_put(info, TAG_LOWER_BOUND_LIST, attribs, no_attribs, data, no_data, 'elem')
         call saoxml_copy_array_to_data2( idd%profiles(profile_index)%plasma_frequency_upper_bounds, &
                                          idd%profiles(profile_index)%num_tabulated )
         call xml_put(info, TAG_UPPER_BOUND_LIST, attribs, no_attribs, data, no_data, 'elem')
         call xml_put(info, TAG_PROFILE_VALUE_LIST, attribs, no_attribs, data, no_data, 'close')
      endif

      if( optional_density_profile_output ) then

      ! plasma densities (even though can be derived from frequencies)
         no_attribs = 3
         attribs(1,1) = ATTR_NAME
         attribs(2,1) = VAL_PLASMA_DENSITY
         attribs(1,2) = ATTR_TYPE
         attribs(2,2) = VAL_FLOAT
         attribs(1,3) = ATTR_UNITS
         attribs(2,3) = 'cm-3'
         call saoxml_copy_array_to_data2( idd%profiles(profile_index)%plasma_densities, &
                                          idd%profiles(profile_index)%num_tabulated )
         call xml_put(info, TAG_PROFILE_VALUE_LIST, attribs, no_attribs, data, no_data, 'elem')
      
         if( idd%profiles(profile_index)%plasma_density_uncertainty_available ) then
            ! uncertainty bounds for densities
            no_attribs = 0
            call saoxml_copy_array_to_data2( idd%profiles(profile_index)%plasma_density_bounds, &
                                             idd%profiles(profile_index)%num_tabulated )
            call xml_put(info, TAG_BOUND_LIST, attribs, no_attribs, data, no_data, 'elem')
            call saoxml_copy_array_to_data2( idd%profiles(profile_index)%plasma_density_lower_bounds, &
                                          idd%profiles(profile_index)%num_tabulated )
            call xml_put(info, TAG_LOWER_BOUND_LIST, attribs, no_attribs, data, no_data, 'elem')
            call saoxml_copy_array_to_data2( idd%profiles(profile_index)%plasma_density_upper_bounds, &
                                             idd%profiles(profile_index)%num_tabulated )
            call xml_put(info, TAG_UPPER_BOUND_LIST, attribs, no_attribs, data, no_data, 'elem')
            call xml_put(info, TAG_PROFILE_VALUE_LIST, attribs, no_attribs, data, no_data, 'close')
         endif
      endif
      call xml_put(info, TAG_TABULATED, attribs, no_attribs, data, no_data, 'close')
   endif

   ! store Chebyshev representation second
   if( idd%profiles(profile_index)%f2_chebyshev%num + &
       idd%profiles(profile_index)%f1_chebyshev%num + &
       idd%profiles(profile_index)%e_chebyshev%num     > 0 ) then
      
      no_attribs = 0
      no_data = 0
      call xml_put(info, TAG_SHIFTED_CHEBYSHEV_LIST, attribs, no_attribs, data, no_data, 'open')
      call saoxml_write_one_chebysev_set( idd%profiles(profile_index)%f2_chebyshev )
      call saoxml_write_one_chebysev_set( idd%profiles(profile_index)%f1_chebyshev )
      call saoxml_write_one_chebysev_set( idd%profiles(profile_index)%e_chebyshev )
      call xml_put(info, TAG_SHIFTED_CHEBYSHEV_LIST, attribs, no_attribs, data, no_data, 'close')
   endif

   ! store quasi-parabolic segments third
   if( idd%profiles(profile_index)%num_qp_segments > 0 ) then
   
      no_attribs=2
      no_data = 0
      attribs(1,1) = ATTR_NUM
      attribs(2,1) = icompact( idd%profiles(profile_index)%num_qp_segments )
      attribs(1,2) = ATTR_EARTH_RADIUS
      attribs(2,2) = compact( idd%profiles(profile_index)%earth_radius )
      call xml_put(info, TAG_QUASI_PARABOLIC_LIST, attribs, no_attribs, data, no_data, 'open')
      do i=1, idd%profiles(profile_index)%num_qp_segments
         no_attribs=7
         no_data = 0
         attribs(1,1) = ATTR_ID
         attribs(2,1) = icompact( idd%profiles(profile_index)%qps(i)%id )
         attribs(1,2) = ATTR_START_DISTANCE
         attribs(2,2) = compact( idd%profiles(profile_index)%qps(i)%start_distance_km )
         attribs(1,3) = ATTR_END_DISTANCE
         attribs(2,3) = compact( idd%profiles(profile_index)%qps(i)%end_distance_km )
         attribs(1,4) = ATTR_A
         attribs(2,4) = compact( idd%profiles(profile_index)%qps(i)%a )
         attribs(1,5) = ATTR_B
         attribs(2,5) = compact( idd%profiles(profile_index)%qps(i)%b )
         attribs(1,6) = ATTR_C
         attribs(2,6) = compact( idd%profiles(profile_index)%qps(i)%c )
         attribs(1,7) = ATTR_ERROR
         attribs(2,7) = compact( idd%profiles(profile_index)%qps(i)%fit_error )
         call xml_put(info, TAG_QUASI_PARABOLIC, attribs, no_attribs, data, no_data, 'elem')
      enddo
      call xml_put(info, TAG_QUASI_PARABOLIC_LIST, attribs, no_attribs, data, no_data, 'close')
   endif

   ! store valley model data next
   if( idd%profiles(profile_index)%valley_width_km .ne. NO_VALUE_CHAR .or. &
       idd%profiles(profile_index)%valley_start_height_km .ne. NO_VALUE_CHAR ) then
      no_attribs=0
      no_data = 0
      if( len_trim(idd%profiles(profile_index)%valley_model) > 0 ) then
         no_attribs = no_attribs + 1
         attribs(1,no_attribs) = ATTR_MODEL
         attribs(2,no_attribs) = trim(idd%profiles(profile_index)%valley_model )
      endif
      if( idd%profiles(profile_index)%valley_width_km .ne. NO_VALUE_CHAR ) then
         no_attribs = no_attribs + 1
         attribs(1,no_attribs) = ATTR_WIDTH
         attribs(2,no_attribs) = compact(idd%profiles(profile_index)%valley_width_km )
      endif
      if( idd%profiles(profile_index)%valley_depth_pfu .ne. NO_VALUE_CHAR ) then
         no_attribs = no_attribs + 1
         attribs(1,no_attribs) = ATTR_DEPTH
         attribs(2,no_attribs) = compact(idd%profiles(profile_index)%valley_depth_pfu )
      endif
      if( idd%profiles(profile_index)%valley_start_height_km .ne. NO_VALUE_CHAR ) then
         no_attribs = no_attribs + 1
         attribs(1,no_attribs) = ATTR_START_HEIGHT
         attribs(2,no_attribs) = compact(idd%profiles(profile_index)%valley_start_height_km )
      endif
      if( idd%profiles(profile_index)%valley_start_freq_MHz .ne. NO_VALUE_CHAR ) then
         no_attribs = no_attribs + 1
         attribs(1,no_attribs) = ATTR_START_FREQUENCY
         attribs(2,no_attribs) = compact(idd%profiles(profile_index)%valley_start_freq_MHz )
      endif
      call xml_put(info, TAG_PROFILE_VALLEY, attribs, no_attribs, data, no_data, 'elem')
   endif
   
   ! store POLAN coefficients - ToDo
   
   ! store topside profile Chapman-layer representation
   if( idd%profiles(profile_index)%top_chap%available ) then
      no_attribs=3
      no_data = 0
      attribs(1,1) = ATTR_PEAK_HEIGHT
      attribs(2,1) = compact( idd%profiles(profile_index)%top_chap%peak_height_km )
      attribs(1,2) = ATTR_PEAK_DENSITY
      attribs(2,2) = compact( idd%profiles(profile_index)%top_chap%peak_density_cm3 )
      attribs(1,3) = ATTR_PEAK_SCALE_HEIGHT
      attribs(2,3) = compact( idd%profiles(profile_index)%top_chap%peak_scale_height_km )
      if( len_trim( idd%profiles(profile_index)%top_chap%model ) > 0 ) then
         no_attribs=no_attribs+1
         attribs(1,no_attribs) = ATTR_MODEL
         attribs(2,no_attribs) = trim(idd%profiles(profile_index)%top_chap%model)
      endif
      call xml_put(info, TAG_TOPSIDE_CHAPMAN, attribs, no_attribs, data, no_data, 'elem')
   endif
   
   ! store algorithm options
   if( len_trim( idd%profiles(profile_index)%algorithm_options ) > 0 ) then
      no_attribs = 0
      no_data = 1
      data(1) = idd%profiles(profile_index)%algorithm_options
      call xml_put(info, TAG_ALGORITHM_OPTIONS, attribs, no_attribs, data, no_data, 'elem')
   endif

   ! close the element
   call xml_put(info, TAG_PROFILE, attribs, no_attribs, data, no_data, 'close')

end subroutine saoxml_write_one_profile

subroutine saoxml_write_one_chebysev_set( chebyshev )
   type(IDD_PROFILE_COEFS) :: chebyshev

   if( chebyshev%num .eq. 0 ) return
   no_attribs = 6
   attribs(1,1) = ATTR_REGION
   attribs(2,1) = trim( chebyshev%region )
   attribs(1,2) = ATTR_START_FREQUENCY
   attribs(2,2) = compact( chebyshev%start_freq_MHz )
   attribs(1,3) = ATTR_END_FREQUENCY
   attribs(2,3) = compact( chebyshev%end_freq_MHz )
   attribs(1,4) = ATTR_PEAK_HEIGHT
   attribs(2,4) = compact( chebyshev%peak_height_km )
   attribs(1,5) = ATTR_ERROR
   attribs(2,5) = compact( chebyshev%fit_error )
   attribs(1,6) = ATTR_NUM
   attribs(2,6) = icompact( chebyshev%num )
   if( chebyshev%z_half_Nm_km .ne. NO_VALUE_CHAR ) then
      no_attribs = 7
      attribs(1,7) = ATTR_Z_HALF_NM
      attribs(2,7) = compact( chebyshev%z_half_Nm_km )
   endif
   call saoxml_copy_array_to_data1( chebyshev%c, chebyshev%num )
   call xml_put(info, TAG_SHIFTED_CHEBYSHEV, attribs, no_attribs, data, no_data, 'elem')
end subroutine saoxml_write_one_chebysev_set
!
!=============        s t r i n g    m a n i p u l a t i o n       ================================
!

character(24) function compact( value ) 
   real(8) :: value
   integer :: k
   write( compact, '(F24.6)' ) value
   do k=len(compact),1,-1
      if( compact(k:k) .eq. '.' ) then
         compact(k:k) = ' '
         exit
      elseif( compact(k:k) .eq. '0' ) then
         compact(k:k) = ' '
      elseif( compact(k:k) .ne. ' ' ) then
         exit
      endif
   enddo
   compact = adjustl(compact)
end function compact

character(10) function icompact( value ) 
   integer :: value
   write( icompact, '(I0)' ) value
end function icompact

subroutine saoxml_copy_array_to_data1( array, length )
   real(8), dimension(:) :: array
   integer               :: length
   integer               :: j, i1, i2
   if( length .eq. 0 ) return
   if( length .gt. size( array ) ) return ! ToDo: flag the error condition
   i1 = 1
   do j = 1, length
      buffer = compact(array(j))
      i2 = i1+len_trim(buffer)-1
      data(1)(i1:i2)= buffer
      data(1)(i2+1:i2+1) = ' '
      i1 = i2+2
   enddo
   do j = i1, len(data(1))
      data(1)(j:j) = ' '
   enddo
   no_data = 1
end subroutine saoxml_copy_array_to_data1

subroutine saoxml_copy_array_to_data2( array, length )
   real(8), dimension(:) :: array
   integer               :: length
   integer               :: j, i1, i2
   i1 = 1
   if( length < 10 ) then
      call saoxml_copy_array_to_data1( array, length )
      return
   endif
   if( length .gt. size( array ) ) return ! ToDo: flag the error condition
   do j = 1, length/2
      buffer = compact(array(j))
      i2 = i1+len_trim(buffer)-1
      data(1)(i1:i2)= buffer
      data(1)(i2+1:i2+1) = ' '
      i1 = i2+2
   enddo
   do j = i1, len(data(1))
      data(1)(j:j) = ' '
   enddo
   i1 = 1
   do j = length/2+1, length
      buffer = compact(array(j))
      i2 = i1+len_trim(buffer)-1
      data(2)(i1:i2)= buffer
      data(2)(i2+1:i2+1) = ' '
      i1 = i2+2
   enddo
   do j = i1, len(data(2))
      data(2)(j:j) = ' '
   enddo
   no_data = 2
end subroutine saoxml_copy_array_to_data2

end module SAOXML5RW
